#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "types.h"
#include "gecko.h"

/* macro defines */

#define SWISSSERVER_VERSION "v0.2 beta win32"
#define ASK_READY           0x15
#define ASK_OPENPATH        0x16
#define ASK_GETENTRY        0x17
#define ASK_SERVEFILE       0x18
#define ASK_LOCKFILE        0x19

/* external variables */

/* None */

/* local variables */

static U8 ANS_READY = 0x25;
static char *curPath = NULL;
static S32 cached_files_num = 0;
static S32 next_serve_num = 0;
static FILE_HANDLE_TYPE cached_files[2048];
static FILE *served_file_fp;
static char served_file[1024];    /* The file we're currently serving to the GC */
static U32 count = 0;

/* global variables */

/* None */

/* local prototypes */

static char *getCurPath(void);
static void cache_path(void);
static void send_file_data(USB_DATA_REQ_TYPE *req);

/*
================================================================
=
= Function: main
=
= Description:
=
= a little porgram to transfer files between pc and gc using
= usb gecko.
=
================================================================ 
*/
int main(int argc, char *argv[])
{
    U8 resp = 0;

    printf ("swissserver " SWISSSERVER_VERSION "\n"
                "coded by emu_kidid for Swiss + USB Gecko (modified for shuriken USB by happy_bunny)\n\n");

    if(gecko_open() != 0)
    {
        fprintf (stderr, "unable to open the device\n");
        exit (EXIT_FAILURE);
    }

    /* make sure we close the port on exit */
    atexit(gecko_close);
    
    for(;;)
    {
        resp = gecko_read_byte();

        switch(resp)
        {
            case ASK_READY:
                printf("Got Ready? (%02X) request from GC %x\n",resp,count);
                gecko_send_byte(&ANS_READY);
                count++;
            break;

            case ASK_OPENPATH:
                printf("Got Open Path (%02X) request from GC\n",resp);
                gecko_send_byte(&ANS_READY);
                /* Read path requested by GC and set it */
                gecko_read(getCurPath(), 4096);
                printf("Got path: %s\n",getCurPath());
                cache_path();
            break;

            case ASK_GETENTRY:
                /* Serve file info out to the GC from the directory the GC asked for */
                gecko_send_byte(&ANS_READY);
                if(next_serve_num < cached_files_num)
                {
                    printf("Sending entry to GC: %s\n",cached_files[next_serve_num].name);
                    gecko_write(&cached_files[next_serve_num], sizeof(FILE_HANDLE_TYPE));
                    next_serve_num++;
                }
                else
                {
                    printf("Sending NULL entry to GC\n");
                    memset(&cached_files[0],0, sizeof(FILE_HANDLE_TYPE));
                    gecko_write(&cached_files[0], sizeof(FILE_HANDLE_TYPE));
                }
            break;

            case ASK_SERVEFILE:
                printf("Got Serve File (%02X) request from GC\n",resp);
                gecko_send_byte(&ANS_READY);
                /* Read file, offset and size requested by GC */
                char filename[1024];
                memset(&filename[0], 0, 1024);
                gecko_read(&filename[0], 1024);
                /* compare, open file.*/ 
                if(strcmp((const char*)&served_file, &filename[0]))
                {
                    if(served_file_fp)
                    {
                        printf("Closed File %s\n",served_file);
                        fclose(served_file_fp);
                        served_file_fp = NULL;
                    }      
                    
                    strcpy((char*)served_file, &filename[0]);
                    served_file_fp = fopen(&served_file[0], "rb");
                    printf("Opened File %s\n",served_file);
                }
            break;

            case ASK_LOCKFILE:
                printf("Got Lock File (%02X) request from GC\n",resp);
                gecko_send_byte(&ANS_READY);
                while(1)
                {
                    /* Read file, offset and size requested by GC */
                    USB_DATA_REQ_TYPE req;
                    memset(&req, 0, sizeof(USB_DATA_REQ_TYPE));
                    gecko_read(&req, sizeof(USB_DATA_REQ_TYPE));

                    if(req.size != 0)
                    {
                        send_file_data(&req);
                    }
                    else
                    {
                        printf("Got Unlock File (%02X) request from GC\n",resp);
                        break;    /* end locked file mode */
                    }
                }
            break;

            case 0x00: 
                gecko_read_byte();
                printf("Unlock called when already unlocked, this is ok.\n");
            break;

            default:
                printf("Unknown reply from GC!\n");
        }
    }

    printf("done.\n");

    return 0;
}

/*
================================================================
=
= Function: getCurPath
=
= Description:
=
= returns the current path
=
================================================================ 
*/
static char *getCurPath(void)
{
    if(curPath == NULL)
    {
        curPath = malloc(4096);
        memset(curPath, 0, 4096);
        strcpy(curPath, ".");
    }
    
    return curPath;
}

/*
================================================================
=
= Function: cache_path
=
= Description:
=
= Cache a paths contents
=
================================================================ 
*/
static void cache_path(void)
{
    struct dirent *ent;
    struct stat fstat;

    DIR *dir = opendir (getCurPath());
    cached_files_num = 0;
    next_serve_num = 0;
    
    if (dir != NULL)
    {
        printf("Current DIR: %s\n",curPath);
        
        while ((ent = readdir(dir)) != NULL)
        {
            /* Skip parent links */
            if(((strlen(ent->d_name)==2) && !strncmp(ent->d_name, "..", 2)) ||
                ((strlen(ent->d_name)==1) && !strncmp(ent->d_name, ".", 1)))
            {
                continue;
            }
            
            char *path = malloc(4096);
            memset(path,0,4096);
            /* create a path to directory entry we want to get information about */ 
            sprintf(path, "%s/%s",curPath,ent->d_name);
            /* get information about the entry pointed to by path */ 
            stat(path, &fstat);
            /* print this information */
            printf ("%s %i [%s]\n", ent->d_name, fstat.st_size, S_ISDIR(fstat.st_mode) ? "DIR":"FILE");
            /* clear an entry in the cached_files array */
            memset(&cached_files[cached_files_num],0,sizeof(FILE_HANDLE_TYPE));
            /* store information about this directory entry within the cached_files array */ 
            sprintf(&cached_files[cached_files_num].name[0],"%s",path);
            cached_files[cached_files_num].size = fstat.st_size;
            cached_files[cached_files_num].fileAttrib = S_ISDIR(fstat.st_mode) ? IS_DIR:IS_FILE;
            /* inc the file entry counter */ 
            cached_files_num++;
            free(path);
         }
         
         closedir (dir);
    }
    else
    {
        fprintf (stderr, "Could not read directory [%s]\n",getCurPath());
        exit(EXIT_FAILURE);
    }
}

/*
================================================================
=
= Function: send_file_data
=
= Description:
=
= send file data to gc
=
================================================================ 
*/
static void send_file_data(USB_DATA_REQ_TYPE *req)
{
    printf("Read File: offset %08X, size %i          \r", req->offset, req->size);
    
    if(served_file_fp != NULL)
    {
        fseek(served_file_fp, req->offset, SEEK_SET);
        /* Read and Send */
        void *buffer = (void*)malloc(req->size);
        fread(buffer, 1, req->size, served_file_fp);
        gecko_write(buffer, req->size);
        free(buffer);
    }
}



